/*******************************************************************************
 * Copyright 2017 Graz University of Technology
 * EAAF-Core Components has been developed in a cooperation between EGIZ,  
 * A-SIT Plus, A-SIT, and Graz University of Technology.
 *
 * Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * https://joinup.ec.europa.eu/news/understanding-eupl-v12
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *  
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
package at.gv.egiz.eaaf.core.impl.idp.auth.modules;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;

import javax.annotation.PostConstruct;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.core.io.Resource;

import at.gv.egiz.eaaf.core.api.idp.auth.modules.AuthModule;
import at.gv.egiz.eaaf.core.api.idp.process.ExecutionContext;
import at.gv.egiz.eaaf.core.api.idp.process.ProcessEngine;
import at.gv.egiz.eaaf.core.impl.idp.process.ProcessDefinitionParserException;

/**
 * This class handles registering modules. The modules are detected either with
 * the ServiceLoader mechanism or via Spring. All detected modules are ranked
 * according to their priority.
 */
public class ModuleRegistration {

	private static ModuleRegistration instance = new ModuleRegistration();

	private List<AuthModule> priorizedModules = new ArrayList<>();

	@Autowired
	private ApplicationContext ctx;

	@Autowired
	private ProcessEngine processEngine;

	private Logger log = LoggerFactory.getLogger(getClass());

	public static ModuleRegistration getInstance() {
		return instance;
	}

	private ModuleRegistration() {
	}

	@PostConstruct
	private void init() {
		// load modules via the ServiceLoader
		initServiceLoaderModules();

		// load modules via Spring
		initSpringModules();

		// order modules according to their priority
		sortModules();
	}
	
	/**
	 * Discovers modules which use the ServiceLoader mechanism.
	 */
	private void initServiceLoaderModules() {
		log.info("Looking for auth modules.");
		ServiceLoader<AuthModule> loader = ServiceLoader.load(AuthModule.class);
		Iterator<AuthModule> modules = loader.iterator();
		while (modules.hasNext()) {
			AuthModule module = modules.next();
			log.info("Detected module {}", module.getClass().getName());
			registerModuleProcessDefinitions(module);
			priorizedModules.add(module);
		}
	}

	/**
	 * Discovers modules which use Spring.
	 */
	private void initSpringModules() {
		log.debug("Discovering Spring modules.");
		Map<String, AuthModule> modules = ctx.getBeansOfType(AuthModule.class);
		for (AuthModule module : modules.values()) {
			registerModuleProcessDefinitions(module);
			priorizedModules.add(module);
		}
	}

	/**
	 * Registers the resource uris for the module.
	 * 
	 * @param module
	 *            the module.
	 */
	private void registerModuleProcessDefinitions(AuthModule module) {
		for (String uri : module.getProcessDefinitions()) {
			Resource resource = ctx.getResource(uri);
			if (resource.isReadable()) {
				log.info("Registering process definition '{}'.", uri);
				try (InputStream processDefinitionInputStream = resource.getInputStream()) {
					processEngine.registerProcessDefinition(processDefinitionInputStream);
				} catch (IOException e) {
					log.error("Process definition '{}' could NOT be read.", uri, e);
				} catch (ProcessDefinitionParserException e) {
					log.error("Error while parsing process definition '{}'", uri, e);
				}
			} else {
				log.error("Process definition '{}' cannot be read.", uri);
			}
		}
	}

	/**
	 * Order the modules in descending order according to their priority.
	 */
	private void sortModules() {
		Collections.sort(priorizedModules, new Comparator<AuthModule>() {
			@Override
			public int compare(AuthModule thisAuthModule, AuthModule otherAuthModule) {
				int thisOrder = thisAuthModule.getPriority();
				int otherOrder = otherAuthModule.getPriority();
				return (thisOrder < otherOrder ? 1 : (thisOrder == otherOrder ? 0 : -1));
			}
		});
	}

	/**
	 * Returns the process description id of the first process, in the highest ranked
	 * module, which is able to work with the given execution context.
	 * 
	 * @param context
	 *            the {@link ExecutionContext}.
	 * @return the process id or {@code null}
	 */
	public String selectProcess(ExecutionContext context) {
		for (AuthModule module : priorizedModules) {
			String id = module.selectProcess(context);
			if (StringUtils.isNotEmpty(id)) {
				log.debug("Process with id '{}' selected, for context '{}'.", id, context);
				return id;
			}
		}
		log.info("No process is able to handle context '{}'.", context);
		return null;
	}
}
