/*******************************************************************************
 * Copyright 2017 Graz University of Technology
 * EAAF-Core Components has been developed in a cooperation between EGIZ,  
 * A-SIT Plus, A-SIT, and Graz University of Technology.
 *
 * Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * https://joinup.ec.europa.eu/news/understanding-eupl-v12
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *  
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
package at.gv.egiz.eaaf.core.impl.idp.process.model;

import java.io.Serializable;

import at.gv.egiz.eaaf.core.impl.idp.process.ProcessDefinitionParser;

/**
 * Represents a single transition from a {@link StartEvent} or {@linkplain TaskInfo Task} to another
 * {@linkplain TaskInfo Task} or {@link EndEvent}.
 * 
 * @author tknall
 * 
 */
public class Transition implements Serializable {

	private static final long serialVersionUID = 1L;

	private String id;
	private String conditionExpression;
	private ProcessNode from;
	private ProcessNode to;

	/**
	 * Returns the process node (effectively a {@link StartEvent} or {@linkplain TaskInfo Task}) the transition is
	 * pointing from.
	 * 
	 * @return The transition's source process node (never {@code null} if transition comes from a process definition
	 *         from {@link ProcessDefinitionParser}).
	 */
	public ProcessNode getFrom() {
		return from;
	}

	/**
	 * Sets the process node the transition is pointing from.
	 * 
	 * @param from
	 *            The transition's source process node.
	 */
	public void setFrom(ProcessNode from) {
		this.from = from;
	}

	/**
	 * Returns the process node (effectively a {@linkplain TaskInfo Task} or {@link EndEvent}) the transition is
	 * pointing to.
	 * 
	 * @return The transition's destination process node (never {@code null} if transition comes from a process
	 *         definition from {@link ProcessDefinitionParser}).
	 */
	public ProcessNode getTo() {
		return to;
	}

	/**
	 * Sets the process node the transition is pointing to.
	 * 
	 * @param to
	 *            The transition's destination process node.
	 */
	public void setTo(ProcessNode to) {
		this.to = to;
	}

	/**
	 * Returns the unique identifier of the transition.
	 * 
	 * @return The unique identifier (may be {@code null}).
	 */
	public String getId() {
		return id;
	}

	/**
	 * Sets the unique identifier of the transition.
	 * 
	 * @param id
	 *            The unique identifier.
	 */
	public void setId(String id) {
		this.id = id;
	}

	/**
	 * Returns the condition expression for this transition.
	 * 
	 * @return The condition expression (may be {@code null}).
	 */
	public String getConditionExpression() {
		return conditionExpression;
	}

	/**
	 * Sets the condition expression for this transition.
	 * 
	 * @param conditionExpression
	 *            The condition expression.
	 */
	public void setConditionExpression(String conditionExpression) {
		this.conditionExpression = conditionExpression;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		if (id != null) {
			builder.append("id=");
			builder.append(id);
		}
		if (from != null) {
			if (builder.length() > 0) {
				builder.append(", ");
			}
			builder.append("from.id=");
			builder.append(from.getId());
		}
		if (to != null) {
			if (builder.length() > 0) {
				builder.append(", ");
			}
			builder.append("to.id=");
			builder.append(to.getId());
		}
		if (conditionExpression != null) {
			if (builder.length() > 0) {
				builder.append(", ");
			}
			builder.append("conditionExpression=");
			builder.append(conditionExpression);
		}
		builder.insert(0, "Transition [");
		builder.append("]");
		return builder.toString();
	}

}
