/*******************************************************************************
 * Copyright 2017 Graz University of Technology
 * EAAF-Core Components has been developed in a cooperation between EGIZ,  
 * A-SIT Plus, A-SIT, and Graz University of Technology.
 *
 * Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * https://joinup.ec.europa.eu/news/understanding-eupl-v12
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *  
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
package at.gv.egiz.eaaf.core.api.idp;

import java.net.URI;
import java.net.URL;
import java.util.Map;

import javax.annotation.Nullable;

import at.gv.egiz.eaaf.core.exceptions.EAAFConfigurationException;
import at.gv.egiz.eaaf.core.exceptions.EAAFException;

public interface IConfiguration {
	
	public static final String CONFIG_PROPS_AUTH_DEFAULT_COUNTRYCODE = "configuration.auth.default.countrycode";
	
	/**
	 * Get a configuration value from file based configuration
	 * 
	 * @param key configuration key 
	 * @return configuration value or null if it is not found
	 */
	public String getBasicConfiguration(final String key);
	
	
	/**
	 * Get a configuration value from file based configuration
	 * 
	 * @param key configuration key 
	 * @param defaultValue Default value if no value with this key is found
	 * @return configuration value 
	 */
	public String getBasicConfiguration(final String key, final String defaultValue);
	
	/**
	 * Get a set of configuration values from file based configuration that starts with this prefix
	 * <br><br>
	 * <b>Important:</b> The configuration values must be of type String! 
	 * 
	 * @param prefix Prefix of the configuration key
	 * @return Map<String, String> without prefix, but never null
	 */
	public Map<String, String> getBasicMOAIDConfigurationWithPrefix(final String prefix);
	
	
	/**
	 * Get a boolean value from basic MOA-ID configuration file
	 * 
	 * @param key Configuration key
	 * @param defaultValue Default result
	 * @return returns the value of the configuration key, or the default value if the key is not set
	 */
	public boolean getBasicMOAIDConfigurationBoolean(String key, boolean defaultValue);
	
	/**
	 * Get a configuration entry for a specific Service Provider 
	 * 
	 * @param uniqueID Unique identifier of the Service Provider
	 * @return {@link ISPConfiguration} or <code>null</code> if no SP configuration was found 
	 * @throws EAAFConfigurationException 
	 */
	@Nullable
	public ISPConfiguration getServiceProviderConfiguration(final String uniqueID) throws EAAFConfigurationException; 
	
	
	/**
	 * Get a configuration entry for a specific Service Provider that is decorated by a Object
	 * 
	 * @param spIdentifier EntityID of a Service Provider 
	 * @param decorator Decorator that should be used to decorate the result. 
  	 *                  This decorator has to be implement or extend the {@link ISPConfiguration} interface
	 * @return <code>T</code> or <code>null</code> if no SP configuration was found 
	 * @throws EAAFConfigurationException
	 */
	@Nullable
    public <T> T getServiceProviderConfiguration(String spIdentifier, final Class<T> decorator) throws EAAFConfigurationException;
		
	/**
	 * Get the root directory of the configuration folder
	 * 
	 * @return 
	 */
	public URI getConfigurationRootDirectory();
	
	/**
	 * Validate a URL if it it is allowed by configuration.
	 * 
	 * @param authReqUrl URL for validation
	 * @return URL of the application context if the authReqUrl was valid, otherwise null 
	 */
	public String validateIDPURL(URL authReqUrl) throws EAAFException;
	
	
}
