/*******************************************************************************
 * Copyright 2017 Graz University of Technology
 * EAAF-Core Components has been developed in a cooperation between EGIZ,  
 * A-SIT Plus, A-SIT, and Graz University of Technology.
 *
 * Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * https://joinup.ec.europa.eu/news/understanding-eupl-v12
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *  
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
/*******************************************************************************
 *******************************************************************************/
package at.gv.egiz.eaaf.core.api.idp;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

import at.gv.egiz.eaaf.core.api.data.EAAFConstants;

public interface ISPConfiguration extends Serializable {

	public static final String CONFIG_KEY_RESTRICTIONS_BASEID_INTERNAL = "configuration.restrictions.baseID.idpProcessing";
	public static final String CONFIG_KEY_RESTRICTIONS_BASEID_TRANSMISSION = "configuration.restrictions.baseID.spTransmission";
	
	public static final String CONFIG_KEY_RESTRICTIONS_EID_DEMO_MODE = "configuration.restrictions.eIDTestMode";
	
	/**
	 * Get the full key/value configuration for this Service Provider
	 * 
	 * @return an unmodifiable map of key/value pairs
	 */
	public Map<String, String> getFullConfiguration();
	
	  /**
	   * Get a configuration value from Service Provider key/value configuration
	   * 
	   * @param key The key identifier of a configuration value 
	   * @return The configuration value {String} or null if the key does not exist
	   */  
	public String getConfigurationValue(String key);

	  /**
	   * Get a configuration value from Service Provider key/value configuration
	   * 
	   * @param key The key identifier of a configuration value 
	   * @param defaultValue Default value if key does not exist
	   * @return The configuration value {String} or defaultValue if the key does not exist
	   */  
	public String getConfigurationValue(String key, String defaultValue);
	
	/**
	 * Get a boolean configuration value from Service Provider key/value configuration
	 * 
	 * @param key The key identifier of a configuration value 
	 * @return true / false, or null if the key does not exist
	 */
	public Boolean isConfigurationValue(String key);
	
	
	/**
	 * Get a boolean configuration value from Service Provider key/value configuration
	 * 
	 * @param key The key identifier of a configuration value 
	 * @param defaultValue Default value if key does not exist
	 * @return true / false, or defaultValue if the key does not exist
	 */
	public boolean isConfigurationValue(String key, boolean defaultValue);
	
	/**
	 * Check if a configuration key is available in this Service Provider configuration 
	 * 
	 * @param key The key identifier of a configuration value
	 * @return true if the configuration key exists, otherwise false
	 */
	boolean containsConfigurationKey(String key);
	
	/**
	 * Return the unique identifier of this Service Provider
	 * 
	 * @return
	 */
	public String getUniqueIdentifier();
	
	/**
	 * Return the unique identifier of this Service Provider
	 * 
	 * @return
	 */
	public String getFriendlyName();
	
	/**
	 * Indicates if this service provider has private area restrictions that disallow baseId processing in general
	 * 	 * 
	 * @return true if there is a restriction, otherwise false
	 */
	public boolean hasBaseIdInternalProcessingRestriction();
	
	
	/**
	 * Indicates if this service provider has private area restrictions that disallow baseId transfer to SP
	 *  
	 * @return true if there is a restriction, otherwise false
	 */
	public boolean hasBaseIdTransferRestriction();
	
	/**
	 * Get the {@link List} of identifier's that indicates no baseID processing restriction exists.<br>  
	 * This list can be configured by key: "configuration.restrictions.baseID.idpProcessing"
	 * 
	 * @return
	 */
	public List<String> getTargetsWithNoBaseIdInternalProcessingRestriction();
	
	/**
	 * Get the {@link List} of identifier's that indicates no baseID transfer restriction exists.<br>  
	 * This list can be configured by key: "configuration.restrictions.baseID.spTransmission"
	 * 
	 * @return
	 */
	public List<String> getTargetsWithNoBaseIdTransferRestriction();
	
	/**
	 * Get the List eIDAS LoA that are required by this service provider
	 * 
	 * {@link EAAFConstants.EIDAS_LOA_LOW}
	 * {@link EAAFConstants.EIDAS_LOA_SUBSTANTIAL}
	 * {@link EAAFConstants.EIDAS_LOA_HIGH}
	 * or any other non-notified LoA scheme
	 * 
	 * @return return List of eIDAS LoAs
	 */
	public List<String> getRequiredLoA();
	
	/**
	 * Get required matching mode for LoAs
	 * {@link EIDAS_LOA_MATCHING_MINIMUM}
	 * {@link EIDAS_LOA_MATCHING_EXACT}
	 * 
	 * @return
	 */
	public String getLoAMatchingMode();
	
	
	/**
	 * Get the full area-identifier for this service provider to calculate the 
	 * area-specific unique person identifier (bPK, wbPK, eIDAS unique identifier, ...). 
	 * This identifier always contains the full prefix 
	 * 
	 * @return area identifier with prefix  
	 */
	public String getAreaSpecificTargetIdentifier();

}
